import { PluginDefaultExport } from "upload-plugin-sdk/types/PluginDefaultExport";
import { TransformationHandler } from "upload-plugin-sdk/types/transform/TransformationHandler";
import { promises as fsAsync } from "fs";
import path from "path";
import {
  PluginInvocation,
  TransformationStep,
  TransformationStepBase,
  TransformationStepEstimation
} from "upload-plugin-sdk/types/PluginInvocation";
import { assertUnreachable } from "upload-plugin-sdk/internal/TypeUtils";
import {
  PluginEstimationResult,
  PluginInvocationResult,
  PluginTransformationResult
} from "upload-plugin-sdk/types/PluginInvocationResult";
import { EstimationHandler } from "upload-plugin-sdk/types/transform/EstimationHandler";
import { TransformationHandlerParamsBase } from "upload-plugin-sdk/types/transform/TransformationHandlerParamsBase";
import { TransformationLocalFileManager } from "upload-plugin-sdk/internal/TransformationLocalFileManager";

const fileManager = new TransformationLocalFileManager();

export function transform<TParams, TEstimation extends PluginEstimationResult = PluginEstimationResult>({
  estimate,
  run
}: {
  estimate?: EstimationHandler<TParams, TEstimation>;
  run: TransformationHandler<TParams, TEstimation>;
}): PluginDefaultExport<TParams, TEstimation> {
  return async (invocation: PluginInvocation<TParams, TEstimation>): Promise<PluginInvocationResult<TEstimation>> => {
    switch (invocation.type) {
      case "TransformationStep":
        return await runTransformStep(run, invocation);
      case "TransformationStepEstimation": {
        const defaultEstimate: PluginEstimationResult = { estimation: undefined };
        const result: TEstimation =
          estimate === undefined
            ? (defaultEstimate as TEstimation)
            : await runTransformStepEstimation(estimate, invocation);
        return result;
      }

      case "PluginHealthCheck":
        return {
          success: true
        };
      default:
        assertUnreachable(invocation);
    }
  };
}

async function runTransformStep<TParams, TEstimation extends PluginEstimationResult>(
  handler: TransformationHandler<TParams, TEstimation>,
  step: TransformationStep<TParams, TEstimation>
): Promise<PluginTransformationResult> {
  const handlerWithEstimation = async (
    request: TransformationHandlerParamsBase<TParams>
    // eslint-disable-next-line @typescript-eslint/no-invalid-void-type
  ): Promise<void | Partial<PluginTransformationResult>> =>
    await handler({
      ...request,
      estimationResult: step.estimationResult
    });

  const resultMaybe = await withTransformationParams("transformation", handlerWithEstimation, step);
  const result: Partial<PluginTransformationResult> = resultMaybe !== undefined ? resultMaybe : {};
  return {
    async: result.async ?? Promise.resolve()
  };
}

async function runTransformStepEstimation<TParams, TEstimation extends PluginEstimationResult>(
  handler: EstimationHandler<TParams, TEstimation>,
  step: TransformationStepEstimation<TParams>
): Promise<TEstimation> {
  return await withTransformationParams("estimation", handler, step);
}

async function withTransformationParams<TParams, TResult>(
  invocationType: string,
  handler: (request: TransformationHandlerParamsBase<TParams>) => Promise<TResult>,
  { paths, step, transformation, download }: TransformationStepBase<TParams>
): Promise<TResult> {
  const logs: string[] = [];

  /**
   * IMPORTANT!
   * - We want to keep S3 L2 costs down.
   * - We want to reduce latency as much as possible.
   * - As such, we don't log anything by default.
   * - Only the plugins must call the 'log' method (i.e. not the SDK), and even then, we try and avoid it:
   *   - Image Plugin never calls 'log'.
   *   - Archive Plugin never calls 'log'.
   */
  const log = (msg: string): void => {
    const fullMsg = `${new Date().toISOString()} | ${step.plugin.packageName}@${step.plugin.packageVersion} | ${msg}`;
    console.log(fullMsg); // For local development.
    logs.push(fullMsg);
  };

  const resolve = fileManager.resolveFilePath(paths);
  const getMetadata = fileManager.getMetadata(paths);
  const setMetadata = fileManager.setMetadata(paths);

  try {
    const result = await handler({
      download,
      log,
      params: step.params,
      paths,
      resolve,
      getMetadata,
      setMetadata,
      step,
      transformation
    });

    return result;
  } finally {
    // Use an apex path -- "/log/" -- as this is less likely to collide with a file name required by a plugin,
    // since they will most likely be writing to named files, rather than apex files.
    const logUrl = "/log/";
    const logLocalFilePath = resolve(logUrl);
    const writeLog = async (): Promise<void> => {
      if (logs.length === 0) {
        return; // Do not create an empty file.
      }

      await fsAsync.mkdir(path.dirname(logLocalFilePath), { recursive: true });
      await fsAsync.appendFile(logLocalFilePath, [...logs, ""].join("\n"));
    };
    await Promise.all([
      writeLog(),
      setMetadata(logUrl, {
        contentType: "text/plain"
      })
    ]);
  }
}
